# -*- coding: utf-8 -*-


from __future__ import unicode_literals


__author__ = "Sergey Arkhipov <serge@aerialsounds.org>"
__version__ = 0, 2, 0


from .core import isitbullshit, raise_for_problem  # NOQA
from .exceptions import ItIsBullshitError  # NOQA
from .structures import WHATEVER, OrSkipped  # NOQA
from .testcase_mixin import IsItBullshitMixin  # NOQA


# silence for pyflakes
assert isitbullshit
assert raise_for_problem

assert ItIsBullshitError

assert WHATEVER
assert OrSkipped

assert IsItBullshitMixin



# -*- coding: utf-8 -*-


from six import text_type

from .core import raise_for_problem, isitbullshit
from .exceptions import ItIsBullshitError


# noinspection PyPep8Naming
class IsItBullshitMixin(object):

    @staticmethod
    def assertBullshit(suspicious, scheme, reason=None):  # noqa pylint: disable=C0103
        if not isitbullshit(suspicious, scheme):
            raise AssertionError(text_type(reason))

    @staticmethod
    def assertNotBullshit(suspicious, scheme, reason=None):  # noqa pylint: disable=C0103
        try:
            raise_for_problem(suspicious, scheme)
        except ItIsBullshitError as err:
            if reason is None:
                reason = err
            raise AssertionError(text_type(reason))



# -*- coding: utf-8 -*-


WHATEVER = Ellipsis
FUNC_TYPE = type(lambda: 1)


class OrSkipped(object):  # pylint: disable=R0903
    __slots__ = ("_scheme",)

    def __init__(self, func):
        self._scheme = func

    @property
    def validator(self):
        return self._scheme



# -*- coding: utf-8 -*-


from __future__ import unicode_literals

import sys
import re

from six import iteritems, string_types

from .exceptions import ItIsBullshitError
from .structures import WHATEVER, OrSkipped, FUNC_TYPE


def isitbullshit(suspicious, scheme):
    try:
        raise_for_problem(suspicious, scheme)
    except ItIsBullshitError:
        return True
    return False


def raise_for_problem(suspicious, scheme):
    if isinstance(scheme, OrSkipped):
        raise TypeError("Scheme could not be OrSkipped here")

    if shallow_check(suspicious, scheme):
        return

    if isinstance(scheme, dict):
        raise_for_dict_problem(suspicious, scheme)
    elif isinstance(scheme, list):
        raise_for_list_problem(suspicious, scheme)
    elif isinstance(scheme, tuple):
        raise_for_tuple_problem(suspicious, scheme)
    elif isinstance(scheme, string_types):
        raise_for_string_problem(suspicious, scheme)
    elif isinstance(scheme, float):
        raise_for_float_problem(suspicious, scheme)
    elif isinstance(scheme, FUNC_TYPE):
        raise_for_callable_problem(suspicious, scheme)
    else:
        raise ItIsBullshitError(suspicious,
                                "Scheme mismatch {0}".format(scheme))


def shallow_check(suspicious, scheme):
    if suspicious == scheme or suspicious is scheme or scheme is WHATEVER:
        return True

    for func in (isinstance, issubclass):
        try:
            if func(suspicious, scheme):
                return True
        except TypeError:
            continue

    return False


def raise_for_dict_problem(suspicious, scheme):
    if not scheme:
        raise TypeError("Dict scheme should contain at least 1 element")

    if not isinstance(suspicious, dict):
        raise ItIsBullshitError(suspicious, "Type mismatch, should be a dict")

    for key, validator in iteritems(scheme):
        original_validator = validator
        if isinstance(validator, OrSkipped):
            original_validator = validator.validator
        if key not in suspicious:
            if not isinstance(validator, OrSkipped):
                raise ItIsBullshitError(suspicious,
                                        "Missed key {0}".format(key))
        else:
            if suspicious[key] is not validator:
                try:
                    raise_for_problem(suspicious[key], original_validator)
                except ItIsBullshitError as err:
                    raise ItIsBullshitError(suspicious, err)


def raise_for_list_problem(suspicious, scheme):
    if not scheme:
        raise TypeError("List scheme should contain at least 1 element")

    if len(scheme) != 1:
        raise TypeError("List scheme should contain only 1 element")

    if not isinstance(suspicious, (tuple, list)):
        raise ItIsBullshitError(suspicious,
                                "Type mismatch, should be a list or a tuple")
    for item in suspicious:
        try:
            raise_for_problem(item, scheme[0])
        except ItIsBullshitError as err:
            raise ItIsBullshitError(suspicious, err)


def raise_for_tuple_problem(suspicious, scheme):
    if not scheme:
        raise TypeError("Tuple scheme should contain at least 1 element")

    error = None
    for type_validator in scheme:
        try:
            raise_for_problem(suspicious, type_validator)
            break
        except ItIsBullshitError as err:
            error = err
    else:
        if error is not None:
            raise ItIsBullshitError(suspicious, error)


def raise_for_string_problem(suspicious, scheme):
    if not isinstance(suspicious, string_types):
        raise ItIsBullshitError(suspicious, "Should be a string")

    # noinspection PyTypeChecker
    if re.match(scheme, suspicious, re.UNICODE) is None:
        raise ItIsBullshitError(suspicious,
                                "Regex mismatch: {0}".format(scheme))


def raise_for_float_problem(suspicious, scheme):
    if not isinstance(suspicious, float):
        raise ItIsBullshitError(suspicious, "Should be a float")

    if abs(suspicious - scheme) >= sys.float_info.epsilon:
        raise ItIsBullshitError(suspicious, "Should be {0}".format(scheme))


def raise_for_callable_problem(suspicious, validator):
    try:
        validator(suspicious)
    except Exception as err:
        raise ItIsBullshitError(suspicious, err)



# -*- coding: utf-8 -*-


from __future__ import unicode_literals


from six import text_type, PY3


class ItIsBullshitError(ValueError):

    def __init__(self, line, cause):
        super(ItIsBullshitError, self).__init__()

        self.line = text_type(line) + ": "
        self.cause = cause

    def to_strings(self, indent=False):
        indentation = " " * 4 if indent else ""

        if isinstance(self.cause, ItIsBullshitError):
            yield indentation + self.line
            for level in self.cause.to_strings(True):
                yield indentation + level
        else:
            yield indentation + self.line + text_type(self.cause)

    def __repr__(self):
        # noinspection PyUnresolvedReferences
        return repr(list(elem.lstrip() for elem in self.to_strings()))

    # noinspection PyTypeChecker
    def __unicode__(self):
        return "\n".join(self.to_strings())

    if PY3:
        def __str__(self):  # noqa
            return self.__unicode__()
    else:
        def __str__(self):  # noqa
            return self.__unicode__().encode("utf-8")
